@@ -1,6 +1,6 @@
 # @file DBI.R
 #
-# Copyright 2020 Observational Health Data Sciences and Informatics
+# Copyright 2021 Observational Health Data Sciences and Informatics
 #
 # This file is part of DatabaseConnector
 # 
@@ -102,7 +102,7 @@ setClass("DatabaseConnectorDbiConnection",
 #'                   server = "localhost/ohdsi",
 #'                   user = "joe",
 #'                   password = "secret")
-#' querySql(conn, "SELECT * FROM cdm_synpuf.person")
+#' querySql(conn, "SELECT * FROM cdm_synpuf.person;")
 #' dbDisconnet(conn)
 #' }
 #'
@@ -155,7 +155,7 @@ setMethod("dbQuoteIdentifier",
               return(DBI::SQL(character()))
             }
             if (any(is.na(x))) {
-              stop("Cannot pass NA to dbQuoteIdentifier()", call. = FALSE)
+              abort("Cannot pass NA to dbQuoteIdentifier()")
             }
             if (nzchar(conn@identifierQuote)) {
               x <- gsub(conn@identifierQuote, paste0(conn@identifierQuote,
@@ -174,7 +174,7 @@ setMethod("dbQuoteString",
               return(DBI::SQL(character()))
             }
             if (any(is.na(x))) {
-              stop("Cannot pass NA to dbQuoteString()", call. = FALSE)
+              abort("Cannot pass NA to dbQuoteString()")
             }
             if (nzchar(conn@stringQuote)) {
               x <- gsub(conn@stringQuote, paste0(conn@stringQuote, conn@stringQuote), x, fixed = TRUE)
@@ -200,10 +200,11 @@ setMethod("dbSendQuery",
           signature("DatabaseConnectorJdbcConnection", "character"),
           function(conn, statement, ...) {
             if (rJava::is.jnull(conn@jConnection))
-              stop("Connection is closed")
+              abort("Connection is closed")
             batchedQuery <- rJava::.jnew("org.ohdsi.databaseConnector.BatchedQuery",
                                          conn@jConnection,
-                                         statement)
+                                         statement,
+                                         conn@dbms)
             result <- new("DatabaseConnectorResult",
                           content = batchedQuery,
                           type = "batchedQuery",
@@ -327,9 +328,9 @@ setMethod("dbSendStatement",
 #' @export
 setMethod("dbGetRowsAffected", "DatabaseConnectorResult", function(res, ...) {
   if (res@type != "rowsAffected") {
-    stop("Object not result of dbSendStatement")
+    abort("Object not result of dbSendStatement")
   }
-  return(.jsimplify(res@content))
+  return(rJava::.jsimplify(res@content))
 })
 
 #' @inherit
@@ -385,7 +386,9 @@ setMethod("dbListTables",
 setMethod("dbExistsTable",
           signature("DatabaseConnectorConnection", "character"),
           function(conn, name, database = NULL, schema = NULL, ...) {
-            stopifnot(length(name) == 1)
+            if (length(name) != 1) {
+              abort("Name should be a single string")
+            }
             tables <- dbListTables(conn, name = name, database = database, schema = schema)
             return(tolower(name) %in% tolower(tables))
           })
