@@ -1,4 +1,4 @@
-COPY @qname
+COPY @sqlTableName
 FROM 's3://@s3RepoName/@pathToFiles/@fileName'
 CREDENTIALS 'aws_access_key_id=@awsAccessKey;aws_secret_access_key=@awsSecretAccessKey'
 gzip
