@@ -1,5 +1,6 @@
 package org.ohdsi.databaseConnector;
 
+import java.nio.ByteBuffer;
 import java.sql.Connection;
 import java.sql.ResultSet;
 import java.sql.ResultSetMetaData;
@@ -16,10 +17,12 @@ public class BatchedQuery {
 	public static int				STRING			= 2;
 	public static int				DATE			= 3;
 	public static int				DATETIME		= 4;
+	public static int				INTEGER64		= 5;
+	public static int				INTEGER			= 6;
 	public static int				FETCH_SIZE		= 2048;
 	private static SimpleDateFormat	DATE_FORMAT		= new SimpleDateFormat("yyyy-MM-dd");
 	private static SimpleDateFormat	DATETIME_FORMAT	= new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
-
+	
 	private Object[]				columns;
 	private int[]					columnTypes;
 	private String[]				columnNames;
@@ -30,7 +33,24 @@ public class BatchedQuery {
 	private ResultSet				resultSet;
 	private Connection				connection;
 	private boolean					done;
-
+	
+	private static double[] convertToInteger64ForR(long[] value) {
+		double[] result = new double[value.length];
+		ByteBuffer byteBuffer = ByteBuffer.allocate(8 * value.length);
+		for (int i = 0; i < value.length; i++)
+			byteBuffer.putLong(value[i]);
+		byteBuffer.flip();
+		for (int i = 0; i < value.length; i++)
+			result[i] = byteBuffer.getDouble();
+		
+		return result;
+	}
+	
+	public static double[] validateInteger64() {
+		long[] values = new long[] { 1, -1, (long) Math.pow(2, 33), (long) Math.pow(-2, 33) };
+		return convertToInteger64ForR(values);
+	}
+	
 	private void reserveMemory() {
 		System.gc();
 		long available = Runtime.getRuntime().maxMemory() - (Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory());
@@ -47,13 +67,17 @@ public class BatchedQuery {
 		for (int columnIndex = 0; columnIndex < columnTypes.length; columnIndex++)
 			if (columnTypes[columnIndex] == NUMERIC)
 				columns[columnIndex] = new double[batchSize];
+			else if (columnTypes[columnIndex] == INTEGER)
+				columns[columnIndex] = new int[batchSize];
+			else if (columnTypes[columnIndex] == INTEGER64)
+				columns[columnIndex] = new long[batchSize];
 			else if (columnTypes[columnIndex] == STRING)
 				columns[columnIndex] = new String[batchSize];
 			else
 				columns[columnIndex] = new String[batchSize];
 		rowCount = 0;
 	}
-
+	
 	private void trySettingAutoCommit(Connection connection, boolean value) throws SQLException {
 		try {
 			connection.setAutoCommit(value);
@@ -61,8 +85,8 @@ public class BatchedQuery {
 			// Do nothing
 		}
 	}
-
-	public BatchedQuery(Connection connection, String query) throws SQLException {
+	
+	public BatchedQuery(Connection connection, String query, String dbms) throws SQLException {
 		this.connection = connection;
 		if (connection.getAutoCommit())
 			trySettingAutoCommit(connection, false);
@@ -76,8 +100,14 @@ public class BatchedQuery {
 		for (int columnIndex = 0; columnIndex < metaData.getColumnCount(); columnIndex++) {
 			columnSqlTypes[columnIndex] = metaData.getColumnTypeName(columnIndex + 1);
 			int type = metaData.getColumnType(columnIndex + 1);
-			if (type == Types.BIGINT || type == Types.DECIMAL || type == Types.DOUBLE || type == Types.FLOAT || type == Types.INTEGER || type == Types.NUMERIC
-					|| type == Types.REAL || type == Types.SMALLINT || type == Types.TINYINT)
+			String className = metaData.getColumnClassName(columnIndex + 1);
+			int precision = metaData.getPrecision(columnIndex + 1);
+//			System.out.println(metaData.getScale(columnIndex + 1));
+			if (type == Types.INTEGER || type == Types.SMALLINT || type == Types.TINYINT || (dbms.equals("oracle") && className.equals("java.math.BigDecimal") && precision != 19))
+				columnTypes[columnIndex] = INTEGER;
+			else if (type == Types.BIGINT || (dbms.equals("oracle") && className.equals("java.math.BigDecimal") && precision == 19))
+				columnTypes[columnIndex] = INTEGER64;
+			else if (type == Types.DECIMAL || type == Types.DOUBLE || type == Types.FLOAT || type == Types.NUMERIC || type == Types.REAL)
 				columnTypes[columnIndex] = NUMERIC;
 			else if (type == Types.DATE)
 				columnTypes[columnIndex] = DATE;
@@ -93,7 +123,7 @@ public class BatchedQuery {
 		done = false;
 		totalRowCount = 0;
 	}
-
+	
 	public void fetchBatch() throws SQLException {
 		rowCount = 0;
 		while (rowCount < batchSize && resultSet.next()) {
@@ -101,7 +131,15 @@ public class BatchedQuery {
 				if (columnTypes[columnIndex] == NUMERIC) {
 					((double[]) columns[columnIndex])[rowCount] = resultSet.getDouble(columnIndex + 1);
 					if (resultSet.wasNull())
-						((double[]) columns[columnIndex])[rowCount] = Double.NaN;
+						((double[]) columns[columnIndex])[rowCount] = Double.NaN;				
+				} else if (columnTypes[columnIndex] == INTEGER64) {
+					((long[]) columns[columnIndex])[rowCount] = resultSet.getLong(columnIndex + 1);
+					if (resultSet.wasNull())
+						((long[]) columns[columnIndex])[rowCount] = Long.MIN_VALUE;
+				} else if (columnTypes[columnIndex] == INTEGER) {
+					((int[]) columns[columnIndex])[rowCount] = resultSet.getInt(columnIndex + 1);
+					if (resultSet.wasNull())
+						((int[]) columns[columnIndex])[rowCount] = Integer.MIN_VALUE;
 				} else if (columnTypes[columnIndex] == STRING)
 					((String[]) columns[columnIndex])[rowCount] = resultSet.getString(columnIndex + 1);
 				else if (columnTypes[columnIndex] == DATE) {
@@ -116,7 +154,7 @@ public class BatchedQuery {
 						((String[]) columns[columnIndex])[rowCount] = null;
 					else
 						((String[]) columns[columnIndex])[rowCount] = DATETIME_FORMAT.format(timestamp);
-
+					
 				}
 			rowCount++;
 		}
@@ -126,7 +164,7 @@ public class BatchedQuery {
 		}
 		totalRowCount += rowCount;
 	}
-
+	
 	public void clear() {
 		try {
 			resultSet.close();
@@ -138,7 +176,7 @@ public class BatchedQuery {
 			e.printStackTrace();
 		}
 	}
-
+	
 	public double[] getNumeric(int columnIndex) {
 		double[] column = ((double[]) columns[columnIndex - 1]);
 		if (column.length > rowCount) {
@@ -148,7 +186,7 @@ public class BatchedQuery {
 		} else
 			return column;
 	}
-
+	
 	public String[] getString(int columnIndex) {
 		String[] column = ((String[]) columns[columnIndex - 1]);
 		if (column.length > rowCount) {
@@ -158,27 +196,47 @@ public class BatchedQuery {
 		} else
 			return column;
 	}
-
+	
+	public int[] getInteger(int columnIndex) {
+		int[] column = ((int[]) columns[columnIndex - 1]);
+		if (column.length > rowCount) {
+			int[] newColumn = new int[rowCount];
+			System.arraycopy(column, 0, newColumn, 0, rowCount);
+			return newColumn;
+		} else
+			return column;
+	}
+	
+	public double[] getInteger64(int columnIndex) {
+		long[] column = ((long[]) columns[columnIndex - 1]);
+		if (column.length > rowCount) {
+			long[] newColumn = new long[rowCount];
+			System.arraycopy(column, 0, newColumn, 0, rowCount);
+			return convertToInteger64ForR(newColumn);
+		} else
+			return convertToInteger64ForR(column);
+	}
+	
 	public boolean isDone() {
 		return done;
 	}
-
+	
 	public boolean isEmpty() {
 		return (rowCount == 0);
 	}
-
+	
 	public int[] getColumnTypes() {
 		return columnTypes;
 	}
-
+	
 	public String[] getColumnSqlTypes() {
 		return columnSqlTypes;
 	}
-
+	
 	public String[] getColumnNames() {
 		return columnNames;
 	}
-
+	
 	public int getTotalRowCount() {
 		return totalRowCount;
 	}
